@@ -4,6 +4,7 @@ use strict;
 use warnings;
 
 use Config;
+use Cwd;
 use Getopt::Long;
 use Pod::Usage;
 use File::Spec;
@@ -46,7 +47,10 @@ GetOptions(\%args, qw(
     static has-libtommath has-libatomic_ops
     has-dyncall has-libffi pkgconfig=s
     build=s host=s big-endian jit! enable-jit
-    prefix=s bindir=s libdir=s mastdir=s make-install asan ubsan valgrind telemeh show-autovect show-autovect-failed:s),
+    prefix=s bindir=s libdir=s mastdir=s
+    relocatable make-install asan ubsan
+    valgrind telemeh show-autovect git-cache-dir=s
+    show-autovect-failed:s),
 
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
     'no-debug|nodebug' => sub { $args{debug} = 0 },
@@ -68,29 +72,25 @@ Please specify another installation target by using --prefix=PATH.
 ENOTTOCWD
 }
 
-# Override default target directories with command line argumets
+# Override default target directories with command line arguments
 my @target_dirs = qw{bindir libdir mastdir};
 for my $target (@target_dirs) {
     $config{$target} = $args{$target} if $args{$target};
 }
 
-if (-d '.git') {
-    print dots("Updating submodules");
-    my $msg = qx{git submodule sync --quiet && git submodule --quiet update --init 2>&1};
-    if ($? >> 8 == 0) { print "OK\n" }
-    else {
-        if ($msg =~ /[']([^']+)[']\s+already exists and is not an empty/) {
-            $folder_to_delete = "\n\nERROR: Cannot update submodule because directory exists and is not empty.\n" .
-            ">>> Please delete the following folder and try again:\n$1\n\n";
-        }
-        softfail("git error: $msg")
-    }
-}
+# Download / Update submodules
+my $code = system($^X, 'tools/update-submodules.pl', Cwd::cwd(), @args);
+exit 1 if $code >> 8 != 0;
 
 # fiddle with flags
 $args{optimize}     = 3 if not defined $args{optimize} or $args{optimize} eq "";
 $args{debug}        = 3 if defined $args{debug} and $args{debug} eq "";
 
+# Relocatability is not supported on AIX and OpenBSD.
+if ( $args{relocatable} && ($^O eq 'aix' || $^O eq 'openbsd') ) {
+    hardfail('Relocatability is not supported on ' . $^O .
+    ".\n    Leave off the --relocatable flag to do a non-relocatable build.");
+}
 
 for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
         has-libatomic_ops asan ubsan valgrind show-vec)) {
@@ -170,10 +170,17 @@ $config{as}           = $config{cc} unless defined $config{as};
 # If we're in macOS, let's verify that the toolchain is consistent.
 if ($^O eq 'darwin') {
     my $gnu_count = 0;
+    my $gnu_toolchain = (exists $args{toolchain}) && ($args{toolchain} eq 'gnu');
+
+    unless ($gnu_toolchain) {
+        # When XCode toolchain is used then force use of XCode's make if
+        # available.
+        $config{make} = '/usr/bin/make' if -x '/usr/bin/make'; 
+    }
 
     # Here are the tools that seem to cause trouble.
     # If you see other ones, please add them to this list.
-    my @check_tools = qw/ar cc ld make/;
+    my @check_tools = qw/ar cc ld/;
     for my $tool (map { `which $_` } @config{@check_tools}) {
         chomp $tool;
         system "grep -b 'gnu' '$tool'"; # Apple utilities don't match `gnu`
@@ -183,20 +190,14 @@ if ($^O eq 'darwin') {
     }
 
     ## For a GNU toolchain, make sure that they're all GNU.
-    if (exists $args{toolchain} &&
-        $args{toolchain} eq 'gnu' &&
-        $gnu_count != scalar @check_tools)
-    {
+    if ($gnu_toolchain && $gnu_count != scalar @check_tools) {
         print "\nNot all tools in the toolchain are GNU. Please correct this and retry.\n"
             . "See README.markdown for more details.\n\n";
         exit -1;
     }
 
     ## Otherwise, make sure that none of them are GNU
-    elsif ((!exists $args{toolchain} ||
-        $args{toolchain} ne 'gnu' ) &&
-        $gnu_count != 0)
-    {
+    elsif (!$gnu_toolchain && $gnu_count != 0) {
         print "\nGNU tools detected, despite this not being a GNU-oriented build.\n"
             ." Please correct this and retry. See README.markdown for more details.\n\n";
         exit -1;
@@ -214,6 +215,11 @@ if ($config{cc} eq 'gcc' && !$config{can_specific_werror}) {
     $config{ccmiscflags} =~ s/^ +$//;
 }
 
+# Disable RETGUARD on OpenBSD, since it breaks the legojit.
+if ($^O eq 'openbsd' && $args{jit} && $config{cc} eq 'clang') {
+	$config{ccmiscflags} .= ' -fno-ret-protector';
+}
+
 # Set the remaining ldmiscflags. Do this after probing for gcc -Werror probe to not miss that change for the linker.
 $config{ldmiscflags}  = $config{ccmiscflags} unless defined $config{ldmiscflags};
 
@@ -248,6 +254,7 @@ if (-e "$config{pkgconfig}") {
 
 # conditionally set include dirs and install rules
 $config{cincludes} = '' unless defined $config{cincludes};
+$config{moar_cincludes} = '' unless defined $config{moar_cincludes};
 $config{lincludes} = '' unless defined $config{lincludes};
 $config{install}   = '' unless defined $config{install};
 if ($args{'has-libuv'}) {
@@ -256,12 +263,12 @@ if ($args{'has-libuv'}) {
     setup_native_library('libuv') if $config{pkgconfig_works};
 }
 else {
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libuv/include'
-                        . ' ' . $defaults{ccinc} . '3rdparty/libuv/src';
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/libuv\n"
-                        . "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/libuv/uv\n"
-                        . "\t\$(CP) 3rdparty/libuv/include/*.h \$(DESTDIR)\$(PREFIX)/include/libuv\n"
-                        . "\t\$(CP) 3rdparty/libuv/include/uv/*.h \$(DESTDIR)\$(PREFIX)/include/libuv/uv\n";
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libuv/include'
+                             . ' ' . $defaults{ccinc} . '3rdparty/libuv/src';
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libuv\"\n"
+                        . "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libuv/uv\"\n"
+                        . "\t\$(CP) 3rdparty/libuv/include/*.h \"\$(DESTDIR)\$(PREFIX)/include/libuv\"\n"
+                        . "\t\$(CP) 3rdparty/libuv/include/uv/*.h \"\$(DESTDIR)\$(PREFIX)/include/libuv/uv\"\n";
 }
 
 if ($args{'has-libatomic_ops'}) {
@@ -270,27 +277,27 @@ if ($args{'has-libatomic_ops'}) {
     setup_native_library('atomic_ops') if $config{pkgconfig_works};
 }
 else {
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libatomicops/src';
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libatomicops/src';
     my $lao             = '$(DESTDIR)$(PREFIX)/include/libatomic_ops';
-    $config{install}   .= "\t\$(MKPATH) $lao/atomic_ops/sysdeps/armcc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/gcc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/hpc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/ibmc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/icc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/loadstore\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/msftc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/sunc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/*.h $lao\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/*.h $lao/atomic_ops\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/*.h $lao/atomic_ops/sysdeps\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/armcc/*.h $lao/atomic_ops/sysdeps/armcc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/gcc/*.h $lao/atomic_ops/sysdeps/gcc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/hpc/*.h $lao/atomic_ops/sysdeps/hpc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/ibmc/*.h $lao/atomic_ops/sysdeps/ibmc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/icc/*.h $lao/atomic_ops/sysdeps/icc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/loadstore/*.h $lao/atomic_ops/sysdeps/loadstore\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/msftc/*.h $lao/atomic_ops/sysdeps/msftc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/sunc/*.h $lao/atomic_ops/sysdeps/sunc\n";
+    $config{install}   .= "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/armcc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/gcc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/hpc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/ibmc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/icc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/loadstore\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/msftc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/sunc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/*.h \"$lao\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/*.h \"$lao/atomic_ops\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/*.h \"$lao/atomic_ops/sysdeps\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/armcc/*.h \"$lao/atomic_ops/sysdeps/armcc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/gcc/*.h \"$lao/atomic_ops/sysdeps/gcc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/hpc/*.h \"$lao/atomic_ops/sysdeps/hpc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/ibmc/*.h \"$lao/atomic_ops/sysdeps/ibmc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/icc/*.h \"$lao/atomic_ops/sysdeps/icc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/loadstore/*.h \"$lao/atomic_ops/sysdeps/loadstore\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/msftc/*.h \"$lao/atomic_ops/sysdeps/msftc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/sunc/*.h \"$lao/atomic_ops/sysdeps/sunc\"\n";
 }
 
 if ($args{'has-libtommath'}) {
@@ -306,9 +313,9 @@ if ($args{'has-libtommath'}) {
     }
 }
 else {
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libtommath';
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/libtommath\n"
-                        . "\t\$(CP) 3rdparty/libtommath/*.h \$(DESTDIR)\$(PREFIX)/include/libtommath\n";
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libtommath';
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libtommath\"\n"
+                        . "\t\$(CP) 3rdparty/libtommath/*.h \"\$(DESTDIR)\$(PREFIX)/include/libtommath\"\n";
 }
 
 if ($args{'has-libffi'}) {
@@ -349,13 +356,22 @@ elsif ($args{'has-dyncall'}) {
 }
 else {
     $config{nativecall_backend} = 'dyncall';
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/dyncall/dynload'
-                        . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncall'
-                        . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncallback';
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/dyncall\n"
-                        . "\t\$(CP) 3rdparty/dyncall/dynload/*.h \$(DESTDIR)\$(PREFIX)/include/dyncall\n"
-                        . "\t\$(CP) 3rdparty/dyncall/dyncall/*.h \$(DESTDIR)\$(PREFIX)/include/dyncall\n"
-                        . "\t\$(CP) 3rdparty/dyncall/dyncallback/*.h \$(DESTDIR)\$(PREFIX)/include/dyncall\n";
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/dyncall/dynload'
+                             . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncall'
+                             . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncallback';
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
+                        . "\t\$(CP) 3rdparty/dyncall/dynload/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
+                        . "\t\$(CP) 3rdparty/dyncall/dyncall/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
+                        . "\t\$(CP) 3rdparty/dyncall/dyncallback/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n";
+}
+
+if ($config{pkgconfig_works} && has_native_library('libzstd')) {
+    setup_native_library('libzstd');
+    $config{heapsnapformat} = 3;
+}
+else {
+    print "did not find libzstd; will not use heap snapshot format version 3\n";
+    $config{heapsnapformat} = 2;
 }
 
 # mangle library names
@@ -365,6 +381,7 @@ $config{ldlibs} = join ' ',
     (map { sprintf $config{ldsys}, $_; } @{$config{syslibs}});
 $config{ldlibs} = ' -lasan ' . $config{ldlibs} if $args{asan} && $^O ne 'darwin' && $config{cc} ne 'clang';
 $config{ldlibs} = ' -lubsan ' . $config{ldlibs} if $args{ubsan} and $^O ne 'darwin';
+$config{ldlibs} = $config{ldlibs} . ' -lzstd ' if $config{heapsnapformat} == 3;
 # macro defs
 $config{ccdefflags} = join ' ', map { $config{ccdef} . $_ } @{$config{defs}};
 
@@ -409,6 +426,7 @@ push @cflags, '-DDEBUG_HELPERS' if $args{debug};
 push @cflags, '-DMVM_VALGRIND_SUPPORT' if $args{valgrind};
 push @cflags, '-DHAVE_TELEMEH' if $args{telemeh};
 push @cflags, '-DWORDS_BIGENDIAN' if $config{be}; # 3rdparty/sha1 needs it and it isnt set on mips;
+push @cflags, '-DMVM_HEAPSNAPSHOT_FORMAT=' . $config{heapsnapformat};
 push @cflags, $ENV{CFLAGS} if $ENV{CFLAGS};
 push @cflags, $ENV{CPPFLAGS} if $ENV{CPPFLAGS};
 $config{cflags} = join ' ', uniq(@cflags);
@@ -417,13 +435,23 @@ $config{cflags} = join ' ', uniq(@cflags);
 my @ldflags = ($config{ldmiscflags});
 push @ldflags, $config{ldoptiflags}  if $args{optimize};
 push @ldflags, $config{lddebugflags} if $args{debug};
-push @ldflags, $config{ldinstflags}       if $args{instrument};
+push @ldflags, $config{ldinstflags}  if $args{instrument};
 push @ldflags, $config{ld_covflags}  if $args{coverage};
-push @ldflags, $config{ldrpath}           if not $args{static} and $config{prefix} ne '/usr';
-push @ldflags, '-fsanitize=address' if $args{asan};
-push @ldflags, $ENV{LDFLAGS}  if $ENV{LDFLAGS};
+if (not $args{static} and $config{prefix} ne '/usr') {
+    push @ldflags, $config{ldrpath_relocatable} if  $args{relocatable};
+    push @ldflags, $config{ldrpath}             if !$args{relocatable};
+}
+push @ldflags, '-fsanitize=address'  if $args{asan};
+push @ldflags, $ENV{LDFLAGS}         if $ENV{LDFLAGS};
 $config{ldflags} = join ' ', @ldflags;
 
+$config{moarshared} = '';
+# Switch shared lib compiler flags in relocatable case.
+if (not $args{static} and $config{prefix} ne '/usr') {
+    $config{moarshared} = $config{moarshared_relocatable}   if  $args{relocatable};
+    $config{moarshared} = $config{moarshared_norelocatable} if !$args{relocatable};
+}
+
 # setup library names
 $config{moarlib} = sprintf $config{lib}, $NAME;
 $config{moardll} = sprintf $config{dll}, $NAME;
@@ -496,13 +524,14 @@ unless (defined $config{jit_obj}) {
 
 
 if ($config{cc} eq 'cl') {
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/msinttypes\n"
-                        . "\t\$(CP) 3rdparty/msinttypes/*.h \$(DESTDIR)\$(PREFIX)/include/msinttypes\n";
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/msinttypes\"\n"
+                        . "\t\$(CP) 3rdparty/msinttypes/*.h \"\$(DESTDIR)\$(PREFIX)/include/msinttypes\"\n";
 }
 
 build::probe::C_type_bool(\%config, \%defaults);
 build::probe::computed_goto(\%config, \%defaults);
 build::probe::pthread_yield(\%config, \%defaults);
+build::probe::check_fn_malloc_trim(\%config, \%defaults);
 if ($^O eq 'aix') {
     build::probe::numbits(\%config, \%defaults);
     $config{ldflags} = join(',', $config{ldflags}, '-bmaxdata:0x80000000')
@@ -516,7 +545,7 @@ my $order = $config{be} ? 'big endian' : 'little endian';
 print "\n", <<TERM, "\n";
         make: $config{make}
      compile: $config{cc} $config{cflags}
-    includes: $config{cincludes}
+    includes: $config{cincludes} $config{moar_cincludes}
         link: $config{ld} $config{ldflags}
         libs: $config{ldlibs}
 
@@ -640,7 +669,7 @@ available make targets.
 TERM2
 
 if (!$failed && $args{'make-install'}) {
-    system($config{make}, 'install');
+    $failed = system($config{make}, 'install') >> 8;
 }
 print $folder_to_delete if $folder_to_delete;
 exit $failed;
@@ -902,6 +931,17 @@ sub setup_native_library {
     }
 }
 
+sub has_native_library {
+    my $library = shift;
+    my $result_exists = `$config{pkgconfig} --exists $library`;
+    if ($? == 0) {
+        return 1;
+    }
+    else {
+        return 0;
+    }
+}
+
 __END__
 
 =head1 SYNOPSIS
@@ -912,17 +952,17 @@ __END__
                    [--toolchain <toolchain>] [--compiler <compiler>]
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--prefix]
+                   [--static] [--prefix <path>] [--relocatable]
                    [--has-libtommath] [--has-sha] [--has-libuv]
                    [--has-libatomic_ops]
                    [--asan] [--ubsan] [--no-jit]
-                   [--telemeh]
+                   [--telemeh] [--git-cache-dir <path>]
 
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--big-endian] [--prefix]
-                   [--make-install]
+                   [--static] [--prefix <path>] [--relocatable]
+                   [--big-endian] [--make-install]
 
 =head2 Use of environment variables
 
@@ -1054,6 +1094,12 @@ builds, the byte order is auto-detected.
 Install files in subdirectory /bin, /lib and /include of the supplied path.
 The default prefix is "install" if this option is not passed.
 
+=item --relocatable
+
+Search for the libmoar library relative to the executable path.
+(On AIX and OpenBSD MoarVM can not be built relocatable, since both OS'
+miss the necessary mechanism to make this work.)
+
 =item --bindir
 
 Install executable files in the supplied path.  The default is
@@ -1098,4 +1144,13 @@ Disable JIT compiler, which is enabled by default to JIT-compile hot frames.
 
 Build support for the fine-grained internal event logger.
 
+=item --git-cache-dir <path>
+
+Use the given path as a git repository cache.
+For example: --git-cache-dir=/home/user/repo/git_cache
+Each repository ('MoarVM' and its submodules) will use a separate subfolder.
+If the subfolder does not exist, it will be cloned. If it exists the
+contained repository will be updated.
+
+
 =back
