@@ -123,6 +123,9 @@ static const MVMContainerSpec code_pair_spec = {
     code_pair_serialize,
     code_pair_deserialize,
     code_pair_can_store,
+    NULL, /* cas */
+    NULL, /* atomic_load */
+    NULL, /* atomic_store */
     0
 };
 
@@ -138,7 +141,7 @@ static void code_pair_set_container_spec(MVMThreadContext *tc, MVMSTable *st) {
 static void code_pair_configure_container_spec(MVMThreadContext *tc, MVMSTable *st, MVMObject *config) {
     CodePairContData *data = (CodePairContData *)st->container_data;
 
-    MVMROOT(tc, config, {
+    MVMROOT2(tc, config, st, {
         MVMString *fetch = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "fetch");
         MVMString *store;
 
@@ -170,8 +173,8 @@ static void native_ref_fetch_i(MVMThreadContext *tc, MVMObject *cont, MVMRegiste
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_INT)
         MVM_exception_throw_adhoc(tc, "This container does not reference a native integer");
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            res->i64 = MVM_nativeref_read_reg_or_lex_i(tc, cont);
+        case MVM_NATIVEREF_LEX:
+            res->i64 = MVM_nativeref_read_lex_i(tc, cont);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             res->i64 = MVM_nativeref_read_attribute_i(tc, cont);
@@ -192,8 +195,8 @@ static void native_ref_fetch_n(MVMThreadContext *tc, MVMObject *cont, MVMRegiste
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_NUM)
         MVM_exception_throw_adhoc(tc, "This container does not reference a native number");
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            res->n64 = MVM_nativeref_read_reg_or_lex_n(tc, cont);
+        case MVM_NATIVEREF_LEX:
+            res->n64 = MVM_nativeref_read_lex_n(tc, cont);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             res->n64 = MVM_nativeref_read_attribute_n(tc, cont);
@@ -214,8 +217,8 @@ static void native_ref_fetch_s(MVMThreadContext *tc, MVMObject *cont, MVMRegiste
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_STR)
         MVM_exception_throw_adhoc(tc, "This container does not reference a native string");
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            res->s = MVM_nativeref_read_reg_or_lex_s(tc, cont);
+        case MVM_NATIVEREF_LEX:
+            res->s = MVM_nativeref_read_lex_s(tc, cont);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             res->s = MVM_nativeref_read_attribute_s(tc, cont);
@@ -260,8 +263,8 @@ static void native_ref_store_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 v
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_INT)
         MVM_exception_throw_adhoc(tc, "This container does not reference a native integer");
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            MVM_nativeref_write_reg_or_lex_i(tc, cont, value);
+        case MVM_NATIVEREF_LEX:
+            MVM_nativeref_write_lex_i(tc, cont, value);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             MVM_nativeref_write_attribute_i(tc, cont, value);
@@ -282,8 +285,8 @@ static void native_ref_store_n(MVMThreadContext *tc, MVMObject *cont, MVMnum64 v
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_NUM)
         MVM_exception_throw_adhoc(tc, "This container does not reference a native number");
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            MVM_nativeref_write_reg_or_lex_n(tc, cont, value);
+        case MVM_NATIVEREF_LEX:
+            MVM_nativeref_write_lex_n(tc, cont, value);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             MVM_nativeref_write_attribute_n(tc, cont, value);
@@ -304,8 +307,8 @@ static void native_ref_store_s(MVMThreadContext *tc, MVMObject *cont, MVMString
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_STR)
         MVM_exception_throw_adhoc(tc, "This container does not reference a native string");
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            MVM_nativeref_write_reg_or_lex_s(tc, cont, value);
+        case MVM_NATIVEREF_LEX:
+            MVM_nativeref_write_lex_s(tc, cont, value);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             MVM_nativeref_write_attribute_s(tc, cont, value);
@@ -367,6 +370,9 @@ static const MVMContainerSpec native_ref_spec = {
     native_ref_serialize,
     native_ref_deserialize,
     native_ref_can_store,
+    NULL, /* cas */
+    NULL, /* atomic_load */
+    NULL, /* atomic_store */
     1
 };
 
@@ -378,6 +384,88 @@ static void native_ref_configure_container_spec(MVMThreadContext *tc, MVMSTable
     /* Nothing to do. */
 }
 
+void *MVM_container_devirtualize_fetch_for_jit(MVMThreadContext *tc, MVMSTable *st, MVMuint16 type) {
+    if (type != MVM_reg_int64)
+        return NULL;
+    if (st->container_spec == &native_ref_spec) {
+        MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
+        switch (repr_data->ref_kind) {
+            case MVM_NATIVEREF_LEX:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_read_lex_i;
+                    case MVM_reg_num64: return MVM_nativeref_read_lex_n;
+                    case MVM_reg_str:   return MVM_nativeref_read_lex_s;
+                }
+                break;
+            case MVM_NATIVEREF_ATTRIBUTE:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_read_attribute_i;
+                    case MVM_reg_num64: return MVM_nativeref_read_attribute_n;
+                    case MVM_reg_str:   return MVM_nativeref_read_attribute_s;
+                }
+                break;
+            case MVM_NATIVEREF_POSITIONAL:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_read_positional_i;
+                    case MVM_reg_num64: return MVM_nativeref_read_positional_n;
+                    case MVM_reg_str:   return MVM_nativeref_read_positional_s;
+                }
+                break;
+            case MVM_NATIVEREF_MULTIDIM:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_read_multidim_i;
+                    case MVM_reg_num64: return MVM_nativeref_read_multidim_n;
+                    case MVM_reg_str:   return MVM_nativeref_read_multidim_s;
+                }
+                break;
+            default:
+                return NULL;
+        }
+    }
+    return NULL;
+}
+
+void *MVM_container_devirtualize_store_for_jit(MVMThreadContext *tc, MVMSTable *st, MVMuint16 type) {
+    if (type != MVM_reg_int64)
+        return NULL;
+    if (st->container_spec == &native_ref_spec) {
+        MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
+        switch (repr_data->ref_kind) {
+            case MVM_NATIVEREF_LEX:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_write_lex_i;
+                    case MVM_reg_num64: return MVM_nativeref_write_lex_n;
+                    case MVM_reg_str:   return MVM_nativeref_write_lex_s;
+                }
+                break;
+            case MVM_NATIVEREF_ATTRIBUTE:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_write_attribute_i;
+                    case MVM_reg_num64: return MVM_nativeref_write_attribute_n;
+                    case MVM_reg_str:   return MVM_nativeref_write_attribute_s;
+                }
+                break;
+            case MVM_NATIVEREF_POSITIONAL:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_write_positional_i;
+                    case MVM_reg_num64: return MVM_nativeref_write_positional_n;
+                    case MVM_reg_str:   return MVM_nativeref_write_positional_s;
+                }
+                break;
+            case MVM_NATIVEREF_MULTIDIM:
+                switch (type) {
+                    case MVM_reg_int64: return MVM_nativeref_write_multidim_i;
+                    case MVM_reg_num64: return MVM_nativeref_write_multidim_n;
+                    case MVM_reg_str:   return MVM_nativeref_write_multidim_s;
+                }
+                break;
+            default:
+                return NULL;
+        }
+    }
+    return NULL;
+}
+
 static const MVMContainerConfigurer NativeRefContainerConfigurer = {
     native_ref_set_container_spec,
     native_ref_configure_container_spec
@@ -390,37 +478,30 @@ static const MVMContainerConfigurer NativeRefContainerConfigurer = {
 /* Adds a container configurer to the registry. */
 void MVM_6model_add_container_config(MVMThreadContext *tc, MVMString *name,
         const MVMContainerConfigurer *configurer) {
-    void *kdata;
     MVMContainerRegistry *entry;
-    size_t klen;
-
-    MVM_HASH_EXTRACT_KEY(tc, &kdata, &klen, name, "add container config needs concrete string");
 
     uv_mutex_lock(&tc->instance->mutex_container_registry);
 
-    HASH_FIND(hash_handle, tc->instance->container_registry, kdata, klen, entry);
+    MVM_HASH_GET(tc, tc->instance->container_registry, name, entry);
 
     if (!entry) {
         entry = MVM_malloc(sizeof(MVMContainerRegistry));
         entry->name = name;
         entry->configurer  = configurer;
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->name,
+            "Container configuration name");
+        MVM_HASH_BIND(tc, tc->instance->container_registry, name, entry);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->hash_handle.key,
+            "Container configuration hash key");
     }
 
-    HASH_ADD_KEYPTR(hash_handle, tc->instance->container_registry, kdata, klen, entry);
-
     uv_mutex_unlock(&tc->instance->mutex_container_registry);
 }
 
 /* Gets a container configurer from the registry. */
 const MVMContainerConfigurer * MVM_6model_get_container_config(MVMThreadContext *tc, MVMString *name) {
-    void *kdata;
     MVMContainerRegistry *entry;
-    size_t klen;
-
-    MVM_HASH_EXTRACT_KEY(tc, &kdata, &klen, name, "get container config needs concrete string");
-
-    HASH_FIND(hash_handle, tc->instance->container_registry, kdata, klen, entry);
+    MVM_HASH_GET(tc, tc->instance->container_registry, name, entry);
     return entry != NULL ? entry->configurer : NULL;
 }
 
@@ -524,3 +605,128 @@ void MVM_6model_container_assign_s(MVMThreadContext *tc, MVMObject *cont, MVMStr
     else
         MVM_exception_throw_adhoc(tc, "Cannot assign to an immutable value");
 }
+
+/* ***************************************************************************
+ * Container atomic operations
+ * ***************************************************************************/
+
+void MVM_6model_container_cas(MVMThreadContext *tc, MVMObject *cont,
+                              MVMObject *expected, MVMObject *value,
+                              MVMRegister *result) {
+    if (IS_CONCRETE(cont)) {
+        MVMContainerSpec const *cs = cont->st->container_spec;
+        if (cs) {
+            if (cs->cas)
+                cs->cas(tc, cont, expected, value, result);
+            else
+                MVM_exception_throw_adhoc(tc,
+                    "A %s container does not know how to do atomic compare and swap",
+                     MVM_6model_get_stable_debug_name(tc, cont->st));
+        }
+        else {
+            MVM_exception_throw_adhoc(tc,
+                "Cannot perform atomic compare and swap on non-container value of type %s",
+                 MVM_6model_get_stable_debug_name(tc, cont->st));
+        }
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "Cannot perform atomic compare and swap on %s type object",
+             MVM_6model_get_stable_debug_name(tc, cont->st));
+    }
+}
+
+MVMObject * MVM_6model_container_atomic_load(MVMThreadContext *tc, MVMObject *cont) {
+    if (IS_CONCRETE(cont)) {
+        MVMContainerSpec const *cs = cont->st->container_spec;
+        if (cs) {
+            if (cs->atomic_load)
+                return cs->atomic_load(tc, cont);
+            else
+                MVM_exception_throw_adhoc(tc,
+                    "A %s container does not know how to do an atomic load",
+                     MVM_6model_get_stable_debug_name(tc, cont->st));
+        }
+        else {
+            MVM_exception_throw_adhoc(tc,
+                "Cannot perform atomic load from a non-container value of type %s",
+                 MVM_6model_get_stable_debug_name(tc, cont->st));
+        }
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "Cannot perform atomic load from %s type object",
+             MVM_6model_get_stable_debug_name(tc, cont->st));
+    }
+}
+
+void MVM_6model_container_atomic_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
+    if (IS_CONCRETE(cont)) {
+        MVMContainerSpec const *cs = cont->st->container_spec;
+        if (cs) {
+            if (cs->atomic_store)
+                cs->atomic_store(tc, cont, value);
+            else
+                MVM_exception_throw_adhoc(tc,
+                    "A %s container does not know how to do an atomic store",
+                     MVM_6model_get_stable_debug_name(tc, cont->st));
+        }
+        else {
+            MVM_exception_throw_adhoc(tc,
+                "Cannot perform atomic store to a non-container value of type %s",
+                 MVM_6model_get_stable_debug_name(tc, cont->st));
+        }
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "Cannot perform atomic store to %s type object",
+             MVM_6model_get_stable_debug_name(tc, cont->st));
+    }
+}
+
+static AO_t * native_ref_as_atomic_i(MVMThreadContext *tc, MVMObject *cont) {
+    if (REPR(cont)->ID == MVM_REPR_ID_NativeRef && IS_CONCRETE(cont)) {
+        MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
+        if (repr_data->primitive_type == MVM_STORAGE_SPEC_BP_INT) {
+            switch (repr_data->ref_kind) {
+                case MVM_NATIVEREF_LEX:
+                    return MVM_nativeref_as_atomic_lex_i(tc, cont);
+                case MVM_NATIVEREF_ATTRIBUTE:
+                    return MVM_nativeref_as_atomic_attribute_i(tc, cont);
+                case MVM_NATIVEREF_POSITIONAL:
+                    return MVM_nativeref_as_atomic_positional_i(tc, cont);
+                case MVM_NATIVEREF_MULTIDIM:
+                    return MVM_nativeref_as_atomic_multidim_i(tc, cont);
+                default:
+                    MVM_exception_throw_adhoc(tc, "Unknown native int reference kind");
+            }
+        }
+    }
+    MVM_exception_throw_adhoc(tc,
+        "Can only do integer atomic operations on a container referencing a native integer");
+}
+
+MVMint64 MVM_6model_container_cas_i(MVMThreadContext *tc, MVMObject *cont,
+                                    MVMint64 expected, MVMint64 value) {
+    return (MVMint64)MVM_cas(native_ref_as_atomic_i(tc, cont), (AO_t)expected, (AO_t)value);
+}
+
+MVMint64 MVM_6model_container_atomic_load_i(MVMThreadContext *tc, MVMObject *cont) {
+    return (MVMint64)MVM_load(native_ref_as_atomic_i(tc, cont));
+}
+
+void MVM_6model_container_atomic_store_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 value) {
+    MVM_store(native_ref_as_atomic_i(tc, cont), value);
+}
+
+MVMint64 MVM_6model_container_atomic_inc(MVMThreadContext *tc, MVMObject *cont) {
+    return (MVMint64)MVM_incr(native_ref_as_atomic_i(tc, cont));
+}
+
+MVMint64 MVM_6model_container_atomic_dec(MVMThreadContext *tc, MVMObject *cont) {
+    return (MVMint64)MVM_decr(native_ref_as_atomic_i(tc, cont));
+}
+
+MVMint64 MVM_6model_container_atomic_add(MVMThreadContext *tc, MVMObject *cont, MVMint64 value) {
+    return (MVMint64)MVM_add(native_ref_as_atomic_i(tc, cont), (AO_t)value);
+}
