@@ -123,14 +123,20 @@ static void at_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
 static void bind_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *key, MVMRegister value, MVMuint16 kind) {
     MVMString      *name  = (MVMString *)key;
     MVMContextBody *body  = (MVMContextBody *)data;
-    MVMFrame       *frame = body->context;
 
     MVMSpeshFrameWalker fw;
     MVMRegister *found;
     MVMuint16 got_kind;
     MVMFrame *found_frame;
-    if (!setup_frame_walker(tc, &fw, body) || !MVM_spesh_frame_walker_get_lex(tc, &fw,
-            name, &found, &got_kind, 1, &found_frame)) {
+    MVMuint32 needs_root = kind == MVM_reg_obj || kind == MVM_reg_str;
+    MVMuint32 missing;
+    if (needs_root)
+        MVM_gc_root_temp_push(tc, (MVMCollectable **)&(value.o));
+    missing = !setup_frame_walker(tc, &fw, body) || !MVM_spesh_frame_walker_get_lex(tc, &fw,
+            name, &found, &got_kind, 1, &found_frame);
+    if (needs_root)
+        MVM_gc_root_temp_pop(tc);
+    if (missing) {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
         MVM_exception_throw_adhoc_free(tc, waste,
@@ -168,7 +174,6 @@ static MVMuint64 elems(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, voi
 
 static MVMint64 exists_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *key) {
     MVMContextBody *body = (MVMContextBody *)data;
-    MVMFrame *frame = body->context;
 
     MVMSpeshFrameWalker fw;
     MVMRegister *found;
@@ -265,8 +270,9 @@ static const MVMREPROps MVMContext_this_repr = {
 static void snapshot_frame_callees(MVMThreadContext *tc, MVMFrame *f) {
     while (f && f->caller) {
         MVMSpeshCandidate *cand = f->caller->spesh_cand;
+        MVMFrameExtra *extra = MVM_frame_extra(tc, f);
+        extra->caller_info_needed = 1;
         if (cand && cand->num_inlines) {
-            MVMFrameExtra *extra = MVM_frame_extra(tc, f);
             if (cand->jitcode) {
                 if (extra->caller_jit_position)
                     return;
