@@ -4,40 +4,58 @@
    #define MIN(x,y) ((x)<(y)?(x):(y))
 #endif
 
-/* A forced 64-bit version of mp_get_long, since on some platforms long is
- * not all that long. */
-static MVMuint64 mp_get_int64(MVMThreadContext *tc, mp_int * a) {
-    int i, bits;
+/* Get a native int64 from an mp_int. */
+static MVMint64 mp_get_int64(MVMThreadContext *tc, mp_int * a) {
     MVMuint64 res;
+    MVMuint64 signed_max = 9223372036854775807ULL;
+    const int bits = mp_count_bits(a);
 
-    if (a->used == 0) {
-         return 0;
+    /* For 64-bit 2's complement numbers the positive max is 2**63-1, which is 63 bits,
+     * but the negative max is -(2**63), which is 64 bits. */
+    if (MP_NEG == a->sign) {
+        if (bits > 64) {
+            MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
+        }
+        ++signed_max;
+    }
+    else {
+        if (bits > 63) {
+            MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
+        }
     }
 
-    bits = mp_count_bits(a);
+    res = mp_get_mag_ull(a);
+
+    if (res > signed_max) {
+        /* The mp_int was bigger than a signed result could be. */
+        MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
+    }
+
+    return MP_NEG == a->sign ? -res : res;
+}
+
+MVMint64 MVM_p6bigint_get_int64(MVMThreadContext *tc, MVMP6bigintBody *body) {
+    return mp_get_int64(tc, body->u.bigint);
+}
+
+/* Get a native uint64 from an mp_int. */
+static MVMuint64 mp_get_uint64(MVMThreadContext *tc, mp_int * a) {
+    const int bits = mp_count_bits(a);
+
     if (bits > 64) {
         MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
     }
 
-    /* get number of digits of the lsb we have to read */
-    i = MIN(a->used,(int)((sizeof(MVMuint64)*CHAR_BIT+DIGIT_BIT-1)/DIGIT_BIT))-1;
-
-    /* get most significant digit of result */
-    res = DIGIT(a,i);
-
-    while (--i >= 0) {
-        res = (res << DIGIT_BIT) | DIGIT(a,i);
-    }
-    return res;
+    return mp_get_mag_ull(a);
 }
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps P6bigint_this_repr;
 
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st  = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st  = MVM_gc_allocate_stable(tc, &P6bigint_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
@@ -59,8 +77,11 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
     MVMP6bigintBody *src_body = (MVMP6bigintBody *)src;
     MVMP6bigintBody *dest_body = (MVMP6bigintBody *)dest;
     if (MVM_BIGINT_IS_BIG(src_body)) {
+        mp_err err;
         dest_body->u.bigint = MVM_malloc(sizeof(mp_int));
-        mp_init_copy(dest_body->u.bigint, src_body->u.bigint);
+        if ((err = mp_init_copy(dest_body->u.bigint, src_body->u.bigint)) != MP_OKAY) {
+            MVM_exception_throw_adhoc(tc, "Error copying one big integer to another: %s", mp_error_to_string(err));
+        }
     }
     else {
         dest_body->u.smallint.flag = src_body->u.smallint.flag;
@@ -68,6 +89,15 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
     }
 }
 
+void MVM_p6bigint_store_as_mp_int(MVMThreadContext *tc, MVMP6bigintBody *body, MVMint64 value) {
+    mp_err err;
+    mp_int *i = MVM_malloc(sizeof(mp_int));
+    if ((err = mp_init_i64(i, value)) != MP_OKAY) {
+        MVM_exception_throw_adhoc(tc, "Error creating a big integer from a native integer (%"PRIi64"): %s", value, mp_error_to_string(err));
+    }
+    body->u.bigint = i;
+}
+
 static void set_int(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 value) {
     MVMP6bigintBody *body = (MVMP6bigintBody *)data;
     if (MVM_IS_32BIT_INT(value)) {
@@ -75,32 +105,14 @@ static void set_int(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *
         body->u.smallint.value = (MVMint32)value;
     }
     else {
-        mp_int *i = MVM_malloc(sizeof(mp_int));
-        mp_init(i);
-        if (value >= 0) {
-            MVM_bigint_mp_set_uint64(i, (MVMuint64)value);
-        }
-        else {
-            MVM_bigint_mp_set_uint64(i, (MVMuint64)-value);
-            mp_neg(i, i);
-        }
-        body->u.bigint = i;
+        MVM_p6bigint_store_as_mp_int(tc, body, value);
     }
 }
 static MVMint64 get_int(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
     MVMP6bigintBody *body = (MVMP6bigintBody *)data;
     if (MVM_BIGINT_IS_BIG(body)) {
         mp_int *i = body->u.bigint;
-        if (MP_LT == mp_cmp_d(i, 0)) {
-            MVMint64 ret;
-            mp_neg(i, i);
-            ret = mp_get_int64(tc, i);
-            mp_neg(i, i);
-            return -ret;
-        }
-        else {
-            return mp_get_int64(tc, i);
-        }
+        return mp_get_int64(tc, i);
     }
     else {
         return body->u.smallint.value;
@@ -114,9 +126,11 @@ static void set_uint(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
         body->u.smallint.value = (MVMint32)value;
     }
     else {
+        mp_err err;
         mp_int *i = MVM_malloc(sizeof(mp_int));
-        mp_init(i);
-        MVM_bigint_mp_set_uint64(i, value);
+        if ((err = mp_init_u64(i, value)) != MP_OKAY) {
+            MVM_exception_throw_adhoc(tc, "Error creating a big integer from a native integer (%"PRIu64"): %s", value, mp_error_to_string(err));
+        }
         body->u.bigint = i;
     }
 }
@@ -124,10 +138,10 @@ static MVMuint64 get_uint(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
     MVMP6bigintBody *body = (MVMP6bigintBody *)data;
     if (MVM_BIGINT_IS_BIG(body)) {
         mp_int *i = body->u.bigint;
-        if (MP_LT == mp_cmp_d(i, 0))
+        if (MP_NEG == i->sign)
             MVM_exception_throw_adhoc(tc, "Cannot unbox negative bigint into native unsigned integer");
         else
-            return mp_get_int64(tc, i);
+            return mp_get_uint64(tc, i);
     }
     else {
         return body->u.smallint.value;
@@ -139,7 +153,7 @@ static void * get_boxed_ref(MVMThreadContext *tc, MVMSTable *st, MVMObject *root
         return data;
 
     MVM_exception_throw_adhoc(tc,
-        "P6bigint representation cannot unbox to other types");
+        "P6bigint representation cannot unbox to representation %s", MVM_repr_get_by_id(tc, repr_id)->name);
 }
 
 
@@ -183,13 +197,18 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
 static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerializationWriter *writer) {
     MVMP6bigintBody *body = (MVMP6bigintBody *)data;
     if (MVM_BIGINT_IS_BIG(body)) {
+        mp_err err;
         mp_int *i = body->u.bigint;
         int len;
         char *buf;
         MVMString *str;
-        mp_radix_size(i, 10, &len);
+        if ((err = mp_radix_size(i, 10, &len)) != MP_OKAY) {
+            MVM_exception_throw_adhoc(tc, "Error calculating the size of a big integer: %s", mp_error_to_string(err));
+        }
         buf = (char *)MVM_malloc(len);
-        mp_toradix(i, buf, 10);
+        if ((err = mp_to_decimal(i, buf, len)) != MP_OKAY) {
+            MVM_exception_throw_adhoc(tc, "Error converting a big integer to a string: %s", mp_error_to_string(err));
+        }
 
         /* len - 1 because buf is \0-terminated */
         str = MVM_string_ascii_decode(tc, tc->instance->VMString, buf, len - 1);
@@ -219,10 +238,15 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
         body->u.smallint.flag = MVM_BIGINT_32_FLAG;
         body->u.smallint.value = MVM_serialization_read_int(tc, reader);
     } else {  /* big int */
+        mp_err err;
         char *buf = MVM_string_ascii_encode(tc, MVM_serialization_read_str(tc, reader), NULL, 0);
         body->u.bigint = MVM_malloc(sizeof(mp_int));
-        mp_init(body->u.bigint);
-        mp_read_radix(body->u.bigint, buf, 10);
+        if ((err = mp_init(body->u.bigint)) != MP_OKAY) {
+            MVM_exception_throw_adhoc(tc, "Error initializing a big integer: %s", mp_error_to_string(err));
+        }
+        if ((err = mp_read_radix(body->u.bigint, buf, 10)) != MP_OKAY) {
+            MVM_exception_throw_adhoc(tc, "Error converting a string to a big integer: %s", mp_error_to_string(err));
+        }
         MVM_free(buf);
     }
 }
@@ -238,10 +262,10 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
 
 /* Initializes the representation. */
 const MVMREPROps * MVMP6bigint_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &P6bigint_this_repr;
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps P6bigint_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     initialize,
