@@ -805,20 +805,3 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
     MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke");
     return result;
 }
-
-#ifdef _WIN32
-DLLib* MVM_nativecall_load_lib(const char* path) {
-    HMODULE module;
-    if (path != NULL) {
-        const int       len   = MultiByteToWideChar(CP_UTF8, 0, path, -1, NULL, 0);
-        wchar_t * const wpath = (wchar_t *)MVM_malloc(len * sizeof(wchar_t));
-                                MultiByteToWideChar(CP_UTF8, 0, path, -1, (LPWSTR)wpath, len);
-        module = LoadLibraryW(wpath);
-        MVM_free(wpath);
-    }
-    else {
-        module = GetModuleHandle(NULL);
-    }
-    return (DLLib*)(module);
-}
-#endif
