@@ -43,7 +43,10 @@ MVMThreadContext * MVM_tc_create(MVMThreadContext *parent, MVMInstance *instance
     /* Allocate temporary big integers. */
     for (i = 0; i < MVM_NUM_TEMP_BIGINTS; i++) {
         tc->temp_bigints[i] = MVM_malloc(sizeof(mp_int));
-        mp_init(tc->temp_bigints[i]);
+        mp_err err;
+        if ((err = mp_init(tc->temp_bigints[i])) != MP_OKAY) {
+            MVM_exception_throw_adhoc(tc, "Error creating a temporary big integer: %s", mp_error_to_string(err));
+        }
     }
 
     /* Initialize frame sequence numbers */
@@ -76,7 +79,13 @@ void MVM_tc_destroy(MVMThreadContext *tc) {
     MVM_spesh_sim_stack_destroy(tc, tc->spesh_sim_stack);
 
     /* Free the nursery and finalization queue. */
+#if MVM_GC_DEBUG >= 3
+    memset(tc->nursery_fromspace, 0xfe, tc->nursery_fromspace_size);
+#endif
     MVM_free(tc->nursery_fromspace);
+#if MVM_GC_DEBUG >= 3
+    memset(tc->nursery_tospace, 0xfe, tc->nursery_tospace_size);
+#endif
     MVM_free(tc->nursery_tospace);
     MVM_free(tc->finalizing);
 
@@ -110,19 +119,31 @@ void MVM_tc_destroy(MVMThreadContext *tc) {
     }
 
     /* Free the thread context itself. */
-    memset(tc, 0, sizeof(MVMThreadContext));
+    memset(tc, 0xfe, sizeof(MVMThreadContext));
     MVM_free(tc);
 }
 
-/* Setting and clearing mutex to release on exception throw. */
+/* Setting and clearing mutex to release on exception throw.
+ * If the LSB of the mutex' address is set, it's not actually a mutex but a
+ * simple flag (an AO_t) that will be cleared on release. */
 void MVM_tc_set_ex_release_mutex(MVMThreadContext *tc, uv_mutex_t *mutex) {
     if (tc->ex_release_mutex)
         MVM_exception_throw_adhoc(tc, "Internal error: multiple ex_release_mutex");
     tc->ex_release_mutex = mutex;
 }
-void MVM_tc_release_ex_release_mutex(MVMThreadContext *tc) {
+void MVM_tc_set_ex_release_atomic(MVMThreadContext *tc, AO_t *flag) {
     if (tc->ex_release_mutex)
-        uv_mutex_unlock(tc->ex_release_mutex);
+        MVM_exception_throw_adhoc(tc, "Internal error: multiple ex_release_mutex");
+    tc->ex_release_mutex = (uv_mutex_t *)((uintptr_t)flag | 1);
+}
+void MVM_tc_release_ex_release_mutex(MVMThreadContext *tc) {
+    if (tc->ex_release_mutex) {
+        if (MVM_UNLIKELY((uintptr_t)tc->ex_release_mutex & 1)) {
+            *((AO_t*)((uintptr_t)tc->ex_release_mutex & ~(uintptr_t)1)) = 0;
+        } else {
+            uv_mutex_unlock(tc->ex_release_mutex);
+        }
+    }
     tc->ex_release_mutex = NULL;
 }
 void MVM_tc_clear_ex_release_mutex(MVMThreadContext *tc) {
