@@ -7,15 +7,11 @@
 typedef struct {
     /* The libuv handle to the socket. */
     uv_udp_t *handle;
-
-    /* Decode stream, for turning bytes into strings. */
-    MVMDecodeStream *ds;
 } MVMIOAsyncUDPSocketData;
 
 /* Info we convey about a read task. */
 typedef struct {
     MVMOSHandle      *handle;
-    MVMDecodeStream  *ds;
     MVMObject        *buf_type;
     int               seq_number;
     MVMThreadContext *tc;
@@ -34,74 +30,122 @@ static void free_on_close_cb(uv_handle_t *handle) {
     MVM_free(handle);
 }
 
+/* XXX this is duplicated from asyncsocket.c; put it in some shared file */
+static void push_name_and_port(MVMThreadContext *tc, struct sockaddr_storage *name, MVMObject *arr) {
+    char addrstr[INET6_ADDRSTRLEN + 1];
+    /* XXX windows support kludge. 64 bit is much too big, but we'll
+     * get the proper data from the struct anyway, however windows
+     * decides to declare it. */
+    MVMuint64 port;
+    MVMObject *host_o;
+    MVMObject *port_o;
+    if (name) {
+        switch (name->ss_family) {
+            case AF_INET6: {
+                uv_ip6_name((struct sockaddr_in6*)name, addrstr, INET6_ADDRSTRLEN + 1);
+                port = ntohs(((struct sockaddr_in6*)name)->sin6_port);
+                break;
+            }
+            case AF_INET: {
+                uv_ip4_name((struct sockaddr_in*)name, addrstr, INET6_ADDRSTRLEN + 1);
+                port = ntohs(((struct sockaddr_in*)name)->sin_port);
+                break;
+            }
+            default:
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+                return;
+                break;
+        }
+        MVMROOT(tc, arr, {
+            port_o = MVM_repr_box_int(tc, tc->instance->boot_types.BOOTInt, port);
+            MVMROOT(tc, port_o, {
+                host_o = (MVMObject *)MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr,
+                        MVM_string_ascii_decode_nt(tc, tc->instance->VMString, addrstr));
+            });
+        });
+    } else {
+        host_o = tc->instance->boot_types.BOOTStr;
+        port_o = tc->instance->boot_types.BOOTInt;
+    }
+    MVM_repr_push_o(tc, arr, host_o);
+    MVM_repr_push_o(tc, arr, port_o);
+}
+
 /* Read handler. */
 static void on_read(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf, const struct sockaddr *addr, unsigned flags) {
     ReadInfo         *ri  = (ReadInfo *)handle->data;
     MVMThreadContext *tc  = ri->tc;
-    MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, ri->work_idx);
+    MVMObject        *arr;
+    MVMAsyncTask     *t;
+
+    /* libuv will call on_read once after all datagram read operations
+     * to "give us back a buffer". in that case, nread and addr are NULL.
+     * This is an artifact of the underlying implementation and we shouldn't
+     * pass it through to the user. */
+
+    if (nread == 0 && addr == NULL)
+        return;
+
+    arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+    t = MVM_io_eventloop_get_active_work(tc, ri->work_idx);
+
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (nread >= 0) {
-        MVMROOT(tc, t, {
-        MVMROOT(tc, arr, {
+        MVMROOT2(tc, t, arr, {
+            MVMArray *res_buf;
+
             /* Push the sequence number. */
             MVMObject *seq_boxed = MVM_repr_box_int(tc,
                 tc->instance->boot_types.BOOTInt, ri->seq_number++);
             MVM_repr_push_o(tc, arr, seq_boxed);
 
-            /* Either need to produce a buffer or decode characters. */
-            if (ri->ds) {
-                MVMString *str;
-                MVMObject *boxed_str;
-                MVM_string_decodestream_add_bytes(tc, ri->ds, buf->base, nread);
-                str = MVM_string_decodestream_get_all(tc, ri->ds);
-                boxed_str = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, str);
-                MVM_repr_push_o(tc, arr, boxed_str);
-            }
-            else {
-                MVMArray *res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
-                res_buf->body.slots.i8 = (MVMint8 *)buf->base;
-                res_buf->body.start    = 0;
-                res_buf->body.ssize    = buf->len;
-                res_buf->body.elems    = nread;
-                MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
-            }
+            /* Produce a buffer and push it. */
+            res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
+            res_buf->body.slots.i8 = (MVMint8 *)buf->base;
+            res_buf->body.start    = 0;
+            res_buf->body.ssize    = buf->len;
+            res_buf->body.elems    = nread;
+            MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
 
-            /* Finally, no error. */
+            /* next, no error. */
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
-        });
+
+            /* and finally, address and port */
+            push_name_and_port(tc, (struct sockaddr_storage *)addr, arr);
         });
     }
     else if (nread == UV_EOF) {
-        MVMROOT(tc, t, {
-        MVMROOT(tc, arr, {
+        MVMROOT2(tc, t, arr, {
             MVMObject *final = MVM_repr_box_int(tc,
                 tc->instance->boot_types.BOOTInt, ri->seq_number);
             MVM_repr_push_o(tc, arr, final);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
-        });
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
         });
         if (buf->base)
             MVM_free(buf->base);
         uv_udp_recv_stop(handle);
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
     }
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
-        MVMROOT(tc, t, {
-        MVMROOT(tc, arr, {
+        MVMROOT2(tc, t, arr, {
             MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
                 tc->instance->VMString, uv_strerror(nread));
             MVMObject *msg_box = MVM_repr_box_str(tc,
                 tc->instance->boot_types.BOOTStr, msg_str);
             MVM_repr_push_o(tc, arr, msg_box);
-        });
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
         });
         if (buf->base)
             MVM_free(buf->base);
         uv_udp_recv_stop(handle);
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
 }
@@ -114,8 +158,7 @@ static void read_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_t
     /* Add to work in progress. */
     ReadInfo *ri  = (ReadInfo *)data;
     ri->tc        = tc;
-    ri->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    ri->work_idx  = MVM_io_eventloop_add_active_work(tc, async_task);
 
     /* Start reading the stream. */
     handle_data = (MVMIOAsyncUDPSocketData *)ri->handle->body.data;
@@ -149,59 +192,19 @@ static void read_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *workli
 
 /* Frees info for a read task. */
 static void read_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
-    if (data) {
-        ReadInfo *ri = (ReadInfo *)data;
-        if (ri->ds)
-            MVM_string_decodestream_destroy(tc, ri->ds);
+    if (data)
         MVM_free(data);
-    }
 }
 
 /* Operations table for async read task. */
 static const MVMAsyncTaskOps read_op_table = {
     read_setup,
     NULL,
+    NULL,
     read_gc_mark,
     read_gc_free
 };
 
-static MVMAsyncTask * read_chars(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
-                                 MVMObject *schedulee, MVMObject *async_type) {
-    MVMAsyncTask *task;
-    ReadInfo    *ri;
-
-    /* Validate REPRs. */
-    if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
-        MVM_exception_throw_adhoc(tc,
-            "asyncreadchars target queue must have ConcBlockingQueue REPR");
-    if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
-        MVM_exception_throw_adhoc(tc,
-            "asyncreadchars result type must have REPR AsyncTask");
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
-    });
-    });
-    });
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
-    task->body.ops  = &read_op_table;
-    ri              = MVM_calloc(1, sizeof(ReadInfo));
-    ri->ds          = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 0);
-    MVM_ASSIGN_REF(tc, &(task->common.header), ri->handle, h);
-    task->body.data = ri;
-
-    /* Hand the task off to the event loop. */
-    MVMROOT(tc, task, {
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-    });
-
-    return task;
-}
-
 static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
                                  MVMObject *schedulee, MVMObject *buf_type, MVMObject *async_type) {
     MVMAsyncTask *task;
@@ -210,11 +213,12 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     /* Validate REPRs. */
     if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
         MVM_exception_throw_adhoc(tc,
-            "asyncreadbytes target queue must have ConcBlockingQueue REPR");
+            "asyncreadbytes target queue must have ConcBlockingQueue REPR (got %s)",
+             MVM_6model_get_stable_debug_name(tc, queue->st));
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncreadbytes result type must have REPR AsyncTask");
-    if (REPR(buf_type)->ID == MVM_REPR_ID_MVMArray) {
+    if (REPR(buf_type)->ID == MVM_REPR_ID_VMArray) {
         MVMint32 slot_type = ((MVMArrayREPRData *)STABLE(buf_type)->REPR_data)->slot_type;
         if (slot_type != MVM_ARRAY_U8 && slot_type != MVM_ARRAY_I8)
             MVM_exception_throw_adhoc(tc, "asyncreadbytes buffer type must be an array of uint8 or int8");
@@ -224,15 +228,9 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     }
 
     /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, buf_type, {
+    MVMROOT4(tc, queue, schedulee, h, buf_type, {
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
     });
-    });
-    });
-    });
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
     task->body.ops  = &read_op_table;
@@ -252,7 +250,6 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
 /* Info we convey about a write task. */
 typedef struct {
     MVMOSHandle      *handle;
-    MVMString        *str_data;
     MVMObject        *buf_data;
     uv_udp_send_t    *req;
     uv_buf_t          buf;
@@ -266,61 +263,48 @@ static void on_write(uv_udp_send_t *req, int status) {
     WriteInfo        *wi  = (WriteInfo *)req->data;
     MVMThreadContext *tc  = wi->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, wi->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, wi->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (status >= 0) {
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMObject *bytes_box = MVM_repr_box_int(tc,
                 tc->instance->boot_types.BOOTInt,
                 wi->buf.len);
             MVM_repr_push_o(tc, arr, bytes_box);
         });
-        });
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
     }
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
-        MVMROOT(tc, arr, {
-        MVMROOT(tc, t, {
+        MVMROOT2(tc, arr, t, {
             MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
                 tc->instance->VMString, uv_strerror(status));
             MVMObject *msg_box = MVM_repr_box_str(tc,
                 tc->instance->boot_types.BOOTStr, msg_str);
             MVM_repr_push_o(tc, arr, msg_box);
         });
-        });
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
-    if (wi->str_data)
-        MVM_free(wi->buf.base);
     MVM_free(wi->req);
+    MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
 }
 
 /* Does setup work for an asynchronous write. */
 static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     MVMIOAsyncUDPSocketData *handle_data;
-    char                 *output;
-    int                   output_size, r;
+    MVMArray                *buffer;
+    char                    *output;
+    int                      output_size, r;
 
     /* Add to work in progress. */
     WriteInfo *wi = (WriteInfo *)data;
     wi->tc        = tc;
-    wi->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    wi->work_idx  = MVM_io_eventloop_add_active_work(tc, async_task);
 
-    /* Encode the string, or extract buf data. */
-    if (wi->str_data) {
-        MVMuint64 output_size_64;
-        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64, 0);
-        output_size = (int)output_size_64;
-    }
-    else {
-        MVMArray *buffer = (MVMArray *)wi->buf_data;
-        output = (char *)(buffer->body.slots.i8 + buffer->body.start);
-        output_size = (int)buffer->body.elems;
-    }
+    /* Extract buf data. */
+    buffer = (MVMArray *)wi->buf_data;
+    output = (char *)(buffer->body.slots.i8 + buffer->body.start);
+    output_size = (int)buffer->body.elems;
 
     /* Create and initialize write request. */
     wi->req           = MVM_malloc(sizeof(uv_udp_send_t));
@@ -351,6 +335,7 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         /* Cleanup handle. */
         MVM_free(wi->req);
         wi->req = NULL;
+        MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
     }
 }
 
@@ -358,7 +343,6 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
 static void write_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *worklist) {
     WriteInfo *wi = (WriteInfo *)data;
     MVM_gc_worklist_add(tc, worklist, &wi->handle);
-    MVM_gc_worklist_add(tc, worklist, &wi->str_data);
     MVM_gc_worklist_add(tc, worklist, &wi->buf_data);
 }
 
@@ -376,55 +360,11 @@ static void write_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
 static const MVMAsyncTaskOps write_op_table = {
     write_setup,
     NULL,
+    NULL,
     write_gc_mark,
     write_gc_free
 };
 
-static MVMAsyncTask * write_str_to(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
-                                   MVMObject *schedulee, MVMString *s, MVMObject *async_type,
-                                   MVMString *host, MVMint64 port) {
-    MVMAsyncTask    *task;
-    WriteInfo       *wi;
-    struct sockaddr *dest_addr;
-
-    /* Validate REPRs. */
-    if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestrto target queue must have ConcBlockingQueue REPR");
-    if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestrto result type must have REPR AsyncTask");
-
-    /* Resolve destination. */
-    dest_addr = MVM_io_resolve_host_name(tc, host, port);
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, s, {
-        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
-    });
-    });
-    });
-    });
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
-    task->body.ops  = &write_op_table;
-    wi              = MVM_calloc(1, sizeof(WriteInfo));
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->handle, h);
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->str_data, s);
-    wi->dest_addr = dest_addr;
-    task->body.data = wi;
-
-    /* Hand the task off to the event loop. */
-    MVMROOT(tc, task, {
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-    });
-
-    return task;
-}
-
 static MVMAsyncTask * write_bytes_to(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
                                      MVMObject *schedulee, MVMObject *buffer, MVMObject *async_type,
                                      MVMString *host, MVMint64 port) {
@@ -439,25 +379,19 @@ static MVMAsyncTask * write_bytes_to(MVMThreadContext *tc, MVMOSHandle *h, MVMOb
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncwritebytesto result type must have REPR AsyncTask");
-    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "asyncwritebytesto requires a native array to read from");
     if (((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_U8
         && ((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_I8)
         MVM_exception_throw_adhoc(tc, "asyncwritebytesto requires a native array of uint8 or int8");
 
-    /* Resolve destination. */
-    dest_addr = MVM_io_resolve_host_name(tc, host, port);
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, buffer, {
+    /* Resolve destination and create async task handle. */
+    MVMROOT4(tc, queue, schedulee, h, buffer, {
+        MVMROOT(tc, async_type, {
+            dest_addr = MVM_io_resolve_host_name(tc, host, port, MVM_SOCKET_FAMILY_UNSPEC, MVM_SOCKET_TYPE_DGRAM, MVM_SOCKET_PROTOCOL_ANY, 0);
+        });
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
     });
-    });
-    });
-    });
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
     task->body.ops  = &write_op_table;
@@ -490,6 +424,7 @@ static const MVMAsyncTaskOps close_op_table = {
     close_perform,
     NULL,
     NULL,
+    NULL,
     NULL
 };
 
@@ -508,23 +443,33 @@ static MVMint64 close_socket(MVMThreadContext *tc, MVMOSHandle *h) {
     return 0;
 }
 
-static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
-    MVMIOAsyncUDPSocketData *data = (MVMIOAsyncUDPSocketData *)d;
-    if (data->ds) {
-        MVM_string_decodestream_destroy(tc, data->ds);
-        data->ds = NULL;
-    }
+static MVMint64 socket_is_tty(MVMThreadContext *tc, MVMOSHandle *h) {
+    MVMIOAsyncUDPSocketData *data   = (MVMIOAsyncUDPSocketData *)h->body.data;
+    uv_handle_t             *handle = (uv_handle_t *)data->handle;
+    return (MVMint64)(handle->type == UV_TTY);
+}
+
+static MVMint64 socket_handle(MVMThreadContext *tc, MVMOSHandle *h) {
+    MVMIOAsyncUDPSocketData *data   = (MVMIOAsyncUDPSocketData *)h->body.data;
+    uv_handle_t             *handle = (uv_handle_t *)data->handle;
+    int        fd;
+    uv_os_fd_t fh;
+
+    uv_fileno(handle, &fh);
+    fd = uv_open_osfhandle(fh);
+    return (MVMint64)fd;
 }
 
 /* IO ops table, populated with functions. */
 static const MVMIOClosable        closable          = { close_socket };
-static const MVMIOAsyncReadable   async_readable    = { read_chars, read_bytes };
-static const MVMIOAsyncWritableTo async_writable_to = { write_str_to, write_bytes_to };
+static const MVMIOAsyncReadable   async_readable    = { read_bytes };
+static const MVMIOAsyncWritableTo async_writable_to = { write_bytes_to };
+static const MVMIOIntrospection   introspection     = { socket_is_tty,
+                                                        socket_handle };
 static const MVMIOOps op_table = {
     &closable,
     NULL,
     NULL,
-    NULL,
     &async_readable,
     NULL,
     &async_writable_to,
@@ -532,9 +477,10 @@ static const MVMIOOps op_table = {
     NULL,
     NULL,
     NULL,
+    &introspection,
     NULL,
     NULL,
-    gc_free
+    NULL
 };
 
 /* Info we convey about a socket setup task. */
@@ -558,20 +504,21 @@ static void setup_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
 
     if (r >= 0) {
         /* UDP handle initialized; wrap it up in an I/O handle and send. */
-        MVMObject    *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
         MVMAsyncTask *t   = (MVMAsyncTask *)async_task;
-        MVM_repr_push_o(tc, arr, t->body.schedulee);
-        MVMROOT(tc, arr, {
+        MVMObject    *arr;
         MVMROOT(tc, t, {
-            MVMOSHandle          *result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
-            MVMIOAsyncUDPSocketData *data   = MVM_calloc(1, sizeof(MVMIOAsyncUDPSocketData));
-            data->handle                 = udp_handle;
-            result->body.ops             = &op_table;
-            result->body.data            = data;
-            MVM_repr_push_o(tc, arr, (MVMObject *)result);
+            arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+            MVM_repr_push_o(tc, arr, t->body.schedulee);
+            MVMROOT(tc, arr, {
+                MVMOSHandle          *result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
+                MVMIOAsyncUDPSocketData *data   = MVM_calloc(1, sizeof(MVMIOAsyncUDPSocketData));
+                data->handle                 = udp_handle;
+                result->body.ops             = &op_table;
+                result->body.data            = data;
+                MVM_repr_push_o(tc, arr, (MVMObject *)result);
+            });
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
         });
-        });
-        MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
         MVM_repr_push_o(tc, t->body.queue, arr);
     }
     else {
@@ -581,15 +528,13 @@ static void setup_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
             MVMAsyncTask *t   = (MVMAsyncTask *)async_task;
             MVM_repr_push_o(tc, arr, t->body.schedulee);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTIO);
-            MVMROOT(tc, arr, {
-            MVMROOT(tc, t, {
+            MVMROOT2(tc, arr, t, {
                 MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
                     tc->instance->VMString, uv_strerror(r));
                 MVMObject *msg_box = MVM_repr_box_str(tc,
                     tc->instance->boot_types.BOOTStr, msg_str);
                 MVM_repr_push_o(tc, arr, msg_box);
             });
-            });
             MVM_repr_push_o(tc, t->body.queue, arr);
             uv_close((uv_handle_t *)udp_handle, free_on_close_cb);
         });
@@ -611,6 +556,7 @@ static const MVMAsyncTaskOps setup_op_table = {
     setup_setup,
     NULL,
     NULL,
+    NULL,
     setup_gc_free
 };
 
@@ -632,15 +578,16 @@ MVMObject * MVM_io_socket_udp_async(MVMThreadContext *tc, MVMObject *queue,
             "asyncudp result type must have REPR AsyncTask");
 
     /* Resolve hostname. (Could be done asynchronously too.) */
-    if (host && IS_CONCRETE(host))
-        bind_addr = MVM_io_resolve_host_name(tc, host, port);
+    if (host && IS_CONCRETE(host)) {
+        MVMROOT3(tc, queue, schedulee, async_type, {
+            bind_addr = MVM_io_resolve_host_name(tc, host, port, MVM_SOCKET_FAMILY_UNSPEC, MVM_SOCKET_TYPE_DGRAM, MVM_SOCKET_PROTOCOL_ANY, 1);
+        });
+    }
 
     /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
+    MVMROOT2(tc, queue, schedulee, {
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
     });
-    });
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
     MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
     task->body.ops  = &setup_op_table;
