@@ -127,8 +127,10 @@ static void wval_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_ori
 static void wvalfrom_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_orig,
                            MVMuint16 tgt_i, MVMuint16 sslot, MVMint64 idx) {
     MVMSerializationContext *sc = (MVMSerializationContext *)g->spesh_slots[sslot];
-    MVMObject *target = MVM_sc_get_object(tc, sc, idx);
-    object_facts(tc, g, tgt_orig, tgt_i, MVM_sc_try_get_object(tc, sc, idx));
+    if (MVM_sc_is_object_immediately_available(tc, sc, idx)) {
+        MVM_sc_get_object(tc, sc, idx);
+        object_facts(tc, g, tgt_orig, tgt_i, MVM_sc_try_get_object(tc, sc, idx));
+    }
 }
 
 /* Let's figure out what exact type of iter we'll get from an iter op */
@@ -242,7 +244,6 @@ static void sp_guard_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *b
      * justconc and justtype, which don't have a spesh slot. */
     MVMuint16 sslot = ins->operands[ins->info->num_operands - 2].lit_i16;
 
-    MVMSpeshFacts *srcfacts = &g->facts[ins->operands[1].reg.orig][ins->operands[1].reg.i];
     MVMSpeshFacts *facts    = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
 
     /* We do not copy facts here, because it caused some trouble.
@@ -376,7 +377,7 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         guard->operands[1] = preguard_reg;
         guard->operands[2].lit_i16 = MVM_spesh_add_spesh_slot_try_reuse(tc, g,
             (MVMCollectable *)agg_type->st);
-        guard->operands[3].lit_ui32 = g->deopt_addrs[2 * deopt_one_ann->data.deopt_idx];
+        guard->operands[3].lit_ui32 = deopt_one_ann->data.deopt_idx;
         if (ins->next)
             MVM_spesh_manipulate_insert_ins(tc, bb, ins, guard);
         else
@@ -431,7 +432,7 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
 /* Visits the blocks in dominator tree order, recursively. */
 static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                          MVMSpeshPlanned *p) {
-    MVMint32 i, is_phi;
+    MVMint32 i;
 
     /* Look for instructions that provide or propagate facts. */
     MVMSpeshIns *ins = bb->first_ins;
@@ -494,7 +495,6 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 break;
             }
         case MVM_OP_coerce_sI: {
-                MVMSpeshFacts *target_facts = &(g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i]);
                 create_facts(tc, g,
                     ins->operands[0].reg.orig, ins->operands[0].reg.i,
                     ins->operands[2].reg.orig, ins->operands[2].reg.i);
@@ -757,7 +757,7 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
  * instructions that install the block eliminated. This tweaks the usage of
  * them. */
 static void tweak_block_handler_usage(MVMThreadContext *tc, MVMSpeshGraph *g) {
-    MVMint32 i;
+    MVMuint32 i;
     for (i = 0; i < g->sf->body.num_handlers; i++) {
         if (g->sf->body.handlers[i].action == MVM_EX_ACTION_INVOKE) {
             MVMSpeshOperand operand;
