@@ -162,7 +162,7 @@ sub main {
   skip_most:
     break_property('Word', 'Word_Break');
     tweak_nfg_qc();
-
+    find_quick_prop_data();
     # Allocate all the things
     progress("done.\nsetting next_point for codepoints");
     my $first_point = set_next_points();
@@ -199,9 +199,42 @@ sub main {
         write_file("ucd2c_extents.log", $LOG);
     }
     print "\nDONE!!!\n\n";
+    print "Make sure you update tests in roast by following docs/unicode-generated-tests.asciidoc in the roast repo\n";
     return 1;
 }
-
+sub find_quick_prop_data {
+    my @wanted_val_str = (
+        [ 'gencat_name', 'Zl', 'Zp' ],
+    );
+    my @wanted_val_bool = (
+        [ 'White_Space', 1 ]
+    );
+    my %gencat_wanted_h;
+    my @result;
+    for my $code (sort { $a <=> $b } keys %{$POINTS_BY_CODE}) {
+        for my $cat_data (@wanted_val_str) {
+            my $propname = $cat_data->[0];
+            my $i;
+            for ($i = 1; $i < @$cat_data; $i++) {
+                my $pval = $cat_data->[$i];
+                push @{$gencat_wanted_h{$propname . "_" . $pval}}, $code if $POINTS_BY_CODE->{$code}->{$propname} eq $pval;
+            }
+        }
+        for my $cat_data (@wanted_val_bool) {
+            my $propname = $cat_data->[0];
+            push @{$gencat_wanted_h{$propname}}, $code if $POINTS_BY_CODE->{$code}->{$propname};
+        }
+    }
+    say Dumper(%gencat_wanted_h);
+    for my $pname (sort keys %gencat_wanted_h) {
+        my @text;
+        for my $cp (@{$gencat_wanted_h{$pname}}) {
+            push @text, "((cp) == $cp)";
+        }
+        push @result, ("#define MVM_CP_is_$pname(cp) (" . join(' || ', @text) . ')');
+    }
+    write_file("src/strings/unicode_prop_macros.h", (join("\n", @result) . "\n"));
+}
 sub thousands {
     my $in = shift;
     $in = reverse "$in"; # stringify or copy the string
@@ -240,9 +273,27 @@ sub stack_lines {
 }
 
 sub join_sections {
+    # %prefs is a list of sections that need to come before others
+    # The values are sections that need to come before the key.
+    # So: C => [ 'A', 'B' ] would result in A, B, C;
+    my %prefs = (
+        MVM_unicode_get_property_int => ['block_lookup']
+    );
+    my %done;
     my ($sections) = @_;
     my $content = "";
-    $content .= "\n".$sections->{$_} for (sort keys %{$sections});
+    for my $sec (sort keys %{$sections}) {
+        if ($prefs{$sec}) {
+            for my $sec_before (@{$prefs{$sec}}) {
+                next if $done{$sec_before};
+                $content .= "\n".$sections->{$sec_before};
+                $done{$sec_before} = 1;
+            }
+        }
+        next if $done{$sec};
+        $content .= "\n".$sections->{$sec};
+        $done{$sec} = 1;
+    }
     return $content;
 }
 
@@ -752,17 +803,17 @@ static MVMint32 MVM_codepoint_to_row_index(MVMThreadContext *tc, MVMint64 codepo
 
     MVMint32 plane = codepoint >> 16;
 
-    if (codepoint < 0) {
+    if (MVM_UNLIKELY(codepoint < 0)) {
         MVM_exception_throw_adhoc(tc,
             "Internal Error: MVM_codepoint_to_row_index call requested a synthetic codepoint that does not exist.\n"
             "Requested synthetic %%"PRId64" when only %%"PRId32" have been created.",
             -codepoint, tc->instance->nfg->num_synthetics);
     }
 
-    if (plane == 0) {%s
+    if (MVM_LIKELY(plane == 0)) {%s
     }
     else {
-        if (plane < 0 || plane > 16 || codepoint > 0x10FFFD) {
+        if (MVM_UNLIKELY(plane < 0 || plane > 16 || codepoint > 0x10FFFD)) {
             return -1;
         }
         else {%s
@@ -859,22 +910,19 @@ sub emit_property_value_lookup {
     chomp(my $int_out = <<'END');
 
 static MVMint32 MVM_unicode_get_property_int(MVMThreadContext *tc, MVMint64 codepoint, MVMint64 property_code) {
-    MVMuint32 switch_val = (MVMuint32)property_code;
     MVMint32 result_val = 0; /* we'll never have negatives, but so */
-    MVMuint32 codepoint_row = MVM_codepoint_to_row_index(tc, codepoint);
+    MVMint32 codepoint_row = MVM_codepoint_to_row_index(tc, codepoint);
     MVMuint16 bitfield_row;
     /* If codepoint is not found in bitfield rows */
     if (codepoint_row == -1) {
         /* Unassigned codepoints have General Category Cn. Since this returns 0
          * for unknowns, unless we return 1 for property C then these unknows
          * won't match with <:C> */
-        if (property_code == MVM_UNICODE_PROPERTY_C)
-            return 1;
-        return 0;
+        return property_code == MVM_UNICODE_PROPERTY_C ? 1 : 0;
     }
     bitfield_row = codepoint_bitfield_indexes[codepoint_row];
 
-    switch (switch_val) {
+    switch (MVM_EXPECT(property_code, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY)) {
         case 0: return 0;
 END
 
@@ -887,9 +935,18 @@ static const char *bogus = "<BOGUS>"; /* only for table too short; return null s
 static const char* MVM_unicode_get_property_str(MVMThreadContext *tc, MVMint64 codepoint, MVMint64 property_code) {
     MVMuint32 switch_val = (MVMuint32)property_code;
     MVMint32 result_val = 0; /* we'll never have negatives, but so */
-    MVMuint32 codepoint_row = MVM_codepoint_to_row_index(tc, codepoint);
+    MVMint32 codepoint_row;
     MVMuint16 bitfield_row = 0;
 
+    if (switch_val == MVM_UNICODE_PROPERTY_BLOCK) {
+        MVMint32 ord = codepoint;
+        struct UnicodeBlock *block = bsearch(&ord, unicode_blocks, sizeof(unicode_blocks) / sizeof(struct UnicodeBlock), sizeof(struct UnicodeBlock), block_compare);
+        int num = ((char*)block - (char*)unicode_blocks) / sizeof(struct UnicodeBlock);
+        if (block) {
+            return block ? Block_enums[num+1] : Block_enums[0];
+        }
+    }
+    codepoint_row = MVM_codepoint_to_row_index(tc, codepoint);
     if (codepoint_row == -1) { /* non-existent codepoint; XXX should throw? */
         if (0x10FFFF < codepoint)
             return "";
@@ -1114,6 +1171,7 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
         if (codepoint_table_index >= MVM_CODEPOINT_NAMES_COUNT)
             continue;
         switch (codepoint_extents[extent_index][1]) {
+            /* Fate Normal */
             case $FATE_NORMAL: {
                 MVMint32 extent_span_index = 0;
                 codepoint_table_index = codepoint_extents[extent_index][2];
@@ -1131,9 +1189,13 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
                 }
                 break;
             }
+            /* Fate NULL */
             case $FATE_NULL:
+                /* codepoint's += ends up being unused since it gets reset at the next loop.
+                 * For now leaving the code as it was. */
                 codepoint += length;
                 break;
+            /* Fate Span */
             case $FATE_SPAN: {
                 const char *name = codepoint_names[codepoint_table_index];
                 if (name) {
@@ -1289,12 +1351,17 @@ sub add_unicode_sequence {
             $hex_ords = trim shift @list;
             $type     = trim shift @list;
             $name     = trim shift @list;
+            # Don't process non sequences
+            return if $hex_ords =~ /\.\./;
+            return if $hex_ords !~ / /;
         }
         else {
             $name     = trim shift @list;
             $hex_ords = trim shift @list;
             $type     = 'NamedSequences';
         }
+
+
         #\x{23} => chr 0x24
         # It's possible there could be hex unicode digits. In that case convert
         # to the actual codepoints
@@ -1323,7 +1390,7 @@ sub emit_unicode_sequence_keypairs {
     my $count = 0;
     my $seq_c_hash_str = '';
     my @seq_c_hash_array;
-    my $enum_table;
+    my $enum_table = '';
     my $string_seq = "/* Unicode sequences such as Emoji sequences */\n";
     for my $thing ( sort keys %$named_sequences ) {
         my $seq_name = "uni_seq_$count";
@@ -1338,7 +1405,7 @@ sub emit_unicode_sequence_keypairs {
         $ord_data   =~ s/ , $ //x;
         $string_seq =~ s/ , $ //x;
         $string_seq = $string_seq . "}; " . "/* $thing */ /*" . $named_sequences->{$thing}->{'type'} . " */\n";
-        $enum_table = $enum_table . "$seq_name,\n";
+        $enum_table .= "$seq_name,\n";
         $count++;
         if ( length $seq_c_hash_str > 80 ) {
             push @seq_c_hash_array, $seq_c_hash_str . "\n";
